# Credits: Built on open-source libraries and papers acknowledged in README.md citations.
#
"""DiT (Diffusion Transformer) backbone models."""

from src.models.dit.sit import SiTBackbone, SIT_CONFIGS

__all__ = [
    "SiTBackbone",
    "SIT_CONFIGS",
]
