# Copyright (c) 2026 EarthBridge Team.
"""Pix2PixHD example: paired image-to-image translation with Global Generator."""

from examples.pix2pixhd.train_pix2pixhd import Pix2PixHDConfig, Pix2PixHDTrainer

__all__ = [
    "Pix2PixHDConfig",
    "Pix2PixHDTrainer",
]
