# Copyright (c) 2026 EarthBridge Team.
"""BiBBDM (Bidirectional BBDM) training example."""

from .config import BiBBDMConfig
from .train_bibbdm import BiBBDMTrainer

__all__ = ["BiBBDMConfig", "BiBBDMTrainer"]
